/*-
 * Copyright (c) 2009 OKA Shinichiro.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <aio.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>
#include <err.h>

#include <assert.h>

#include <pthread.h>
#include <boost/thread.hpp>
#include <boost/thread/condition.hpp>

#include "asio.hpp"


#define RING_SIZ 64
#define RING_SIZ_MASK 63

struct ringbuf rbuf[RING_SIZ];

int rbuf_rp = -1; // READɤȤ
int rbuf_wp = 0; // ʹWRITE

struct aiocb aio;
int fd = -1;
struct sigevent sigev;

timer_t timerid;

bool inprogress = false;

boost::mutex thread_sync;
boost::condition thread_state;


#define AIO_READ(fd, ringbuf)					\
do {								\
	memset(&aio, 0, sizeof(struct aiocb));			\
	aio.aio_buf = (ringbuf).data;				\
	aio.aio_nbytes = RING_BUFSIZ;				\
	aio.aio_fildes = (fd);					\
	aio.aio_sigevent.sigev_notify = SIGEV_THREAD;		\
	aio.aio_sigevent.sigev_notify_function = read_done;	\
	aio.aio_sigevent.sigev_value.sival_ptr = &aio;		\
	aio_read(&aio);						\
} while (/* CONSTCOND */ 0)


void
asio_stop()
{
	inprogress = false;
	aio_cancel(fd, &aio);
	timer_delete(timerid);

	boost::mutex::scoped_lock lock(thread_sync);
	thread_state.notify_one();
// fprintf(stderr, "stop: notify\n");
}

static void
read_stop(sigval sigv)
{
	asio_stop();
}

static void
read_done(sigval sigv)
{
	int len = aio_return(&aio);
	if (len == -1) {
		perror("aio_read");
	} else if (len > 0) {
		boost::mutex::scoped_lock lock(thread_sync);
		rbuf[rbuf_wp].len = len;

		int w = (rbuf_wp + 1)&RING_SIZ_MASK; // γǼХåե

		if ((w != (rbuf_rp + 1)&RING_SIZ_MASK) || // READɤȤǤʤ
		    (rbuf_rp < 0 && w != 0)) { // ٤ɤǤ餺γǼ褬ƬǤʤ

// fprintf(stderr, "done: wp%d rp%d  update: w%d\n", rbuf_wp, rbuf_rp, w);

			rbuf_wp = w; // rbuf_wpѹΤϤ
			thread_state.notify_one();
// fprintf(stderr, "done: notify\n");
		} else {
 fprintf(stderr, "done: overflow\n");
		}

		pthread_yield();
	} else {
// fprintf(stderr, "done: wp%d rp%d  read none\n", rbuf_wp, rbuf_rp);
	}

	if (inprogress)
		AIO_READ(fd, rbuf[rbuf_wp]);
}

void
asio_start(int filedes, int timeout)
{
	fd = filedes;
	inprogress = true;

	// create timer
	memset(&sigev, 0, sizeof(sigev));
	sigev.sigev_notify = SIGEV_THREAD;
	sigev.sigev_notify_function = read_stop;

	if (timer_create(CLOCK_REALTIME, &sigev, &timerid) == -1)
		err(EXIT_FAILURE, "timer_create");

	// start timer
	struct itimerspec itval;
	memset(&itval, 0, sizeof(itval));
	itval.it_value.tv_sec = timeout; // one-shot

	if (timer_settime(timerid, 0, &itval, NULL) == -1) // no-flag no-signal
		err(EXIT_FAILURE, "timer_settime");

	// start reading
	assert(rbuf_wp == 0);
	AIO_READ(fd, rbuf[rbuf_wp]);
}

// ɤ߹߽λХåեҤȤ֤
// ҤȤĤʤХ֥å
// λƤNULL֤
struct ringbuf*
asio_getbuf()
{
	if (inprogress) {
		boost::mutex::scoped_lock lock(thread_sync);

		int r = (rbuf_rp + 1)&RING_SIZ_MASK; // ɤȤ

		while (r == rbuf_wp) { // WRITEƤȤɤ߹

// fprintf(stderr, "read: waiting\n");
			thread_state.wait(thread_sync);
// fprintf(stderr, "read: awaking\n");

			if (! inprogress)
				return NULL;
		}

// fprintf(stderr, "read: wp%d rp%d  update: %d\n", rbuf_wp, rbuf_rp, r);

		rbuf_rp = r; // rbuf_rpѹΤϤ

		return &rbuf[rbuf_rp];
	} else {
		return NULL;
	}
}
