/*-
 * Copyright (c) 2009 OKA Shinichiro.
 * All rights reserved.
 * Copyright (c) 2008 FUKAUMI Naoki.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cmath>

#include <sys/ioctl.h>
#include <err.h>
#include <fcntl.h>
#include <inttypes.h>
#include <limits.h>
#include <unistd.h>
#include <getopt.h>

#include <dev/usb/usb.h>
#include <dev/usb/usb_ioctl.h>

#include "Hdus.hpp"


struct Options {
	// flag
	int w_b25;

	// arg
	char* channel;
	char* duration;
	char* outfile;

	Options() : w_b25(0), channel(NULL), duration(NULL), outfile(NULL) {}

	void parse(int argc, char *argv[]);
};

void
Options::parse(int argc, char *argv[])
{
	int option_index = 0;
	struct option long_options[] = {
#ifdef B25
		{"b25", no_argument, &w_b25, 1},
#endif
		{0, 0, 0, 0}
	};

	int c;
	do {
		c = getopt_long(argc, argv, "",
		    long_options, &option_index);
	} while (c >= 0);

	if (argc - optind != 3) {
		fprintf(stderr, "usage: %s"
#ifdef B25
		    " [--b25]"
#endif
		    " channel duration outfile\n",
		    getprogname());
		exit(EXIT_FAILURE);
	}

	channel = argv[optind++];
	duration = argv[optind++];
	outfile = argv[optind++];
}

class Channel
{
public:
	Channel() : _catv(0), _null(0), _xpid_min(0x1fff), _xpid_max(0x1fff) {}

	void parse(const char* channel);

	inline int freq() { return _freq; }
	inline int null() { return _null; }
	inline int xpid_min() { return _xpid_min; }
	inline int xpid_max() { return _xpid_max; }

private:
	int chan2freq(int chan);

	int _catv;
	int _freq;
	int _null;
	int _xpid_min;
	int _xpid_max;
};

void
Channel::parse(const char* channel)
{
	if (*channel == 'X') {
		_null = 1; /* output PAT and NULL packet only */
		_xpid_min = 0x0001;
		_xpid_max = 0x1ffe;
		channel++;
	}
	if (*channel == 'X') {
		_null = 2; /* output XOR table in payload of NULL packet */
		channel++;
	}
	if (*channel == 'C') {
		_catv = 1;
		channel++;
	}
	if (*channel == 'C') {
		_catv = 2; /* frequency for C24-C27 may be shifted by +2MHz */
		channel++;
	}

	int chan = strtoul(channel, NULL, 0);
	_freq = chan2freq(chan);
	if (_freq < 0) {
		err(EXIT_FAILURE, "invalid channel\n");
	}
#if 0
	fprintf(stderr, "chan: %s%d%s\n", (_catv > 0) ? "C" : "", chan,
	    (_null > 0) ? " (PAT/NULL only)" : "");
	fprintf(stderr, "freq: %d.142857 MHz\n", _freq);
#endif
}

int
Channel::chan2freq(int chan)
{
	if (_catv == 0) {
		if (chan < 13 || chan > 62)
			return -1;
		return (473 + (chan - 13) * 6);
	} else {
		if (chan < 13 || chan > 63)
			return -1;
		if (chan < 23)
			return (111 + (chan - 13) * 6);
		else if (_catv == 2 && chan >= 24 && chan <= 27)
			return (233 + (chan - 24) * 6);
		else
			return (225 + (chan - 23) * 6);
	}
}


int
main(int argc, char *argv[])
{
	Options opts;
	opts.parse(argc, argv);

	Channel ch;
	ch.parse(opts.channel);

	int duration = strtoul(opts.duration, NULL, 0);
#if 0
	fprintf(stderr, "duration: %d\n", duration);
#endif

	Hdus hdus;
	try {
		hdus.open();
	} catch (const char* e) {
		err(EXIT_FAILURE, "%s", e);
	}

	hdus.init();
	uint8_t saddr = hdus.asie5606(ch.null());

	hdus.setfreq(ch.freq(), ch.xpid_min(), ch.xpid_max());

	hdus.start();

	hdus.record(duration, opts.outfile, opts.w_b25);

	hdus.stop(saddr);

	hdus.close();
	return EXIT_SUCCESS;
}
